//////////////////////////////////////////////////////////////////////////////////////////////////
// OpenGL Framework Class version 2
// --------------------------------
// Author: Thomas Grimes
//
// Framework inspired by the Microsoft(tm) Direct 3D Framework and my older version
// of my OpenGL class. 
//
// Copyright (c) 2000 - 2001 Thomas Grimes. All Rights Reserved.
//////////////////////////////////////////////////////////////////////////////////////////////////

#ifndef OGLFRAME
#define OGLFRAME

#pragma once

#pragma comment( lib, "opengl32.lib" )
#pragma comment( lib, "glu32.lib"    )
#pragma comment( lib, "glaux.lib"    )
#pragma comment( lib, "nvparse.lib"  )


#include <windows.h>
#include <windowsx.h>
#include <fstream.h>
#include <memory.h>
#include <string.h>

#include <gl\gl.h>
#include <gl\glu.h>
#include <gl\glaux.h>

#include <nvgl\nvparse.h>

/*
#define GLH_EXT_SINGLE_FILE	1
// Some of these files already include the other glh headers
#include <nvgl\glh_extensions.h>
#include <nvgl\glh_array.h>
#include <nvgl\glh_combiners.h>
///#include <nvgl\glh_cube_map.h>
#include <nvgl\glh_mipmaps.h>
///#include <nvgl\glh_text.h>
*/

#define COGL_HINT_FOG						0x100000
#define COGL_HINT_LINE_SMOOTH				0x100001
#define COGL_HINT_PERSPECTIVE_CORRECTION	0x100002
#define COGL_HINT_POINT_SMOOTH				0x100003
#define COGL_HINT_POLYGON_SMOOTH			0x100004

#define COGL_HINT_MODE_FASTEST				0x200000
#define COGL_HINT_MODE_NICEST				0x200001
#define COGL_HINT_MODE_DONT_CARE			0x200002

#define KEY_DOWN(vk_code) ((GetAsyncKeyState(vk_code) & 0x8000) ? 1 : 0 )

#ifndef GL_CLAMP_TO_EDGE_EXT
	#define GL_CLAMP_TO_EDGE_EXT	0x812F
#endif


enum RESOLUTIONS
{
	OGL_RES_320x240 = 0,		// Bit 0
	OGL_RES_400x300,			// Bit 1
	OGL_RES_512x384,			// Bit 2
	OGL_RES_640x480,			// Bit 3
	OGL_RES_800x600,			// Bit 4
	OGL_RES_1024x768,			// Bit 5
	OGL_RES_1152x864,			// Bit 6
	OGL_RES_1280x1024,			// Bit 7
	OGL_RES_1600x1200,			// Bit 8
	OGL_RES_2048x1536,			// Bit 9
	// WideScreen mode
	OGL_RES_WS_856x480			// Bit 10
};

enum OGL_ERRORS
{
	OGL_OK = 0,
	OGL_ERR_RELEASE_DC_AND_RC_FAILED,
	OGL_ERR_RELEASE_DC_FAILED,
	OGL_ERR_RELEASE_RC_FAILED,
	OGL_ERR_RELEASE_HWND_FAILED,
	OGL_ERR_UNREGISTER_WNDCLASS_FAILED,
	OGL_ERR_RESOLUTION_NOT_SUPPORTED,
	OGL_ERR_CREATE_DC_FAILED,
	OGL_ERR_CREATE_RC_FAILED,
	OGL_ERR_CANT_FIND_PIXELFORMAT,
	OGL_ERR_CANT_SET_PIXELFORMAT,
	OGL_ERR_ACTIVATE_RC_FAILED,
	OGL_ERR_INIT_FAILED,
	OGL_ERR_CANT_SWAP_BUFFERS,
	OGL_ERR_OPENGL_STATE_INIT_FAILED,
	OGL_ERR_WINDOWED_NOT_SUPPORTED,
	OGL_ERR_COULD_NOT_LOAD_VERTEX_PROGRAM
};

enum WINDOW_MODE
{
	WM_FULLSCREEN = 0,
	WM_WINDOW
};

struct OGLSettings
{
	int bpp;
	int depthbuffer;
	int stencilbuffer;
	int alphabuffer;
	int fov;
	int width;
	int height;
	
	bool window;
};

#define RES_DEFAULT 0
	
// Needed since the window procedure function has to be a global function
LRESULT CALLBACK WndProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam );

class COGLApplication
{
	
protected:

	// Internal variables for the state of the app

	// Used by bit manipulation (maybe)
	// unsigned char = 2 bytes ( 16 bits )
	// 11 current resolutions
	// LSB = lowest res (320x240)
	// Rest of resolutions go from there
	unsigned char m_ucCurrentResolution;
	
	WINDOW_MODE m_wmCurrentWindowMode;

	bool m_bWindowed;
	bool m_bActive;
	bool m_bReady;

	// Internal variables used for timing
	bool m_bFrameMoving;
	bool m_bSingleStep;

	// Error handling
	OGL_ERRORS DisplayErrorMsg( DWORD type );

	// Internal functions to manage and render a 3D scene
	OGL_ERRORS InitializeOpenGLScene( void );
	OGL_ERRORS ToggleFullscreen( void );
	OGL_ERRORS ForceWindowed( void );
	OGL_ERRORS SetupForNewDispSettings( void );
	
	// Calls the virtual function Render and also does other rendering tasks like
	// calculate framerate, check if 3d environment is paused, etc...
	OGL_ERRORS RenderOpenGLScene( void );

//////////////////////////////////////////////////////////////////////////////////////
// Main objects to create and render the scene
//////////////////////////////////////////////////////////////////////////////////////
	HDC		m_hDC;				// Device Context
	HGLRC	m_hRC;				// OpenGL Rendering Context

	HINSTANCE m_hInstance;

	PIXELFORMATDESCRIPTOR m_pfdPixelFormat;

	HWND m_hWnd;				// App window
	HWND m_hWndFocus;		// Uusually same as hWnd
	
	DWORD m_dwWindowStyle;

	RECT m_rWindowBounds;		// Saved window bounds for mode switches
	RECT m_rWindowClient;		// Saved client area size for mode switches

	// Variables for timing
	float m_fTime;				// Current time in seconds
	float m_fElapsedTime;		// Time elapsed since last frame
	float m_fFramesPerSecond;	// Current FPS

	// Overridable variables for the app
	char *m_strWindowTitle;
	char *m_strVtxProgram;
	char *m_strTexShader;
	char *m_strRegCombiner;

	bool m_bUseDepthBuffer;
	bool m_bShowCursor;
	
	int m_iMinDepthBits;
	int m_iMinStencilBits;
	int m_iWindowWidth;
	int m_iWindowHeight;
	int m_iOldWindowWidth;
	int m_iOldWindowHeight;

	GLuint *m_pProgID;
	
	OGLSettings m_oglSettings;

	// Flags to tell if vertex programs, texture shaders, or combiners has been intialized
	bool m_bVtxProgReady;
	bool m_bTexShaderReady;
	bool m_bRegCombinReady;
	bool m_bVidRestart;
	
	// Overridable functions for the scene created by the app
	
	// For things such as glEnable, setting constants for vertex programs, etc...
	virtual OGL_ERRORS InitialSceneInit( void )                   { return OGL_OK; }
	// For loading data such as textures, models, etc... 
	virtual OGL_ERRORS InitObjects( void )                        { return OGL_OK; }
	// Call InitObjects so incase of an error or window change, the objects get reloaded
	virtual OGL_ERRORS RestoreObjects( void )                     { return OGL_OK; }
	// Where the PROJECTION matrix is setup
	virtual OGL_ERRORS ResizeOpenGLScene( int width, int height ) { return OGL_OK; }
	// Used to animate the scene
	virtual OGL_ERRORS NextFrame( void )		                  { return OGL_OK; }
	// Draw every thing here
	virtual OGL_ERRORS Render( void )                             { return OGL_OK; }
	// Release ojbects memeory back to the OS
	virtual OGL_ERRORS CleanupObjects( void )                     { return OGL_OK; }

	virtual void SetupVertexProgram( void );

public:

	// Constructor
	COGLApplication();

	inline HWND GetHwnd( void ) { return m_hWnd; }
	inline HWND GetHwndFocus( void ) { return m_hWndFocus; }

	// Funcs to create, run, pause and cleanup the app
	virtual LRESULT MsgProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam );

	// Main function for initializing opengl
	virtual OGL_ERRORS CreateOpenGLApplication( HINSTANCE hInstance, OGLSettings settings );
	
	// This function is called by the programmer which calls RenderOpenGLScene.
	// That function will call the virtual function Render which the programmer will
	// implement.
	virtual int RunOpenGLApplication( void );

	virtual void PauseOpenGLScene( bool bPause );

	// Other funcs that can not be overidden
	OGL_ERRORS ChangeResolution( RESOLUTIONS res, WINDOW_MODE winMode );
	OGL_ERRORS CleanUpOpenGLScene( void );

	// Returns -1 if one or more extentions doesnt exist
	// Else returns 0 if successful
	virtual int InitExtentions( const char *ext );	

	// Next two funcs will use nvparse()
	void SetTextureShader( const char *shader );
	void SetRegisterCombinerScript( const char *rc );

	bool LoadTextureShaderFromFile( const char *filename );
	bool LoadRegisterCombinerFromFile( const char *filename );
	bool LoadVertexProgramFromFile( const char *filename );

	int SetVertexProgram( GLuint id );		// Returns error in program if any

	inline int GetBitsPerPixel( void )  { return m_oglSettings.bpp;         }
	inline int GetDepthBuffer( void )   { return m_oglSettings.depthbuffer; }
	inline int GetWindowWidth( void )	{ return m_iWindowWidth;          }
	inline int GetWindowHeight( void )	{ return m_iWindowHeight;         }

	inline HDC   GetDeviceContext( void )    { return m_hDC; }
	inline HGLRC GetRenderingContext( void ) { return m_hRC; }

	//inline WINDOW_MODE COGLApplication::GetScreenMode( void ) { return m_wmCurrentWindowMode; }

	void SetHints( int type, int mode );

	virtual void MainGameLoop( void ) {}

	inline char *GetVertexProgram( void ) { return m_strVtxProgram; }

};

#endif